from .type_utils import str2bool

translation = {"type":{str:"STRING", int: "INT", float:"DOUBLE", str2bool:"BOOLEAN", None:"ANY"},
                "column_types": {str:"STRING", None:"ANY", int:"NUMBER", float: "NUMBER"}
                }




from .translation_dictionary import translation
from .type_utils import str2bool


class Property:
    
    def __init__(self, name, type=str, help='', default=None, choices=None, required=False, *args, **kw) -> None:
        
        # Remove all dashes from name
        while name[0] == "-":
            name = name[1:]
        self.name = name

        self.description = help
        self.type = type
        self.default = default
        self.choices = choices
        self.required = required

    def print(self):

        print("#####################")
        print("Name: " + self.name)
        print("Description: ", self.description)
        print("Type: ", translation['type'][self.type])
        print("Default: ", self.default)



{
    "name": "{{ name }}",
    "description": "{{ description }}",
    "mode": "{{mode.upper()}}",
    "metrics": [],
    "area": "{{area.upper()}}",
    "url": "docker://gitlab.alidalab.it:5000/alida/bd-management-services/spark-scdf:3.0.3",
    "version": "1.0.0",
    "framework": {
        "id": 5,
        "name": "Spark",
        "version": "3.2.0"
    },
    "properties": [
        {% for property in properties %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "description": {{json.dumps(property.description)}},
                "mandatory": {{json.dumps(property.required)}},
                "defaultValue": {{json.dumps(property.default)}},
                "value": null,
                "key": {{json.dumps(property.name)}},
                "type": {{json.dumps(translation['type'][property.type])}},
                "inputData": null,
                "outputData": null
            }
        },
        {% endfor %}
        {% for input_dataset in input_datasets %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "defaultValue": null,
                "description": {{json.dumps(input_dataset.description)}},
                "key": "input-dataset",
                "mandatory": true,
                "type": "STRING",
                "value": null,
                "inputData": true,
                "outputData": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "defaultValue": {{json.dumps(translation['column_types'][input_dataset.columns_type])}},
                "description": "Selected columns from table",
                "key": "input-columns",
                "mandatory": true,
                "type": "STRING",
                "value": null,
                "inputData": null,
                "outputData": null
            }
        },
        {% endfor %}
        {% for output_dataset in output_datasets %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "defaultValue": null,
                "description": {{json.dumps(output_dataset.description)}},
                "key": "output-dataset",
                "mandatory": true,
                "type": "STRING",
                "value": null,
                "inputData": null,
                "outputData": true
            }
        },
        {% endfor %}

        {% if input_models|length + output_models|length != 0  %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "The URL for Web HDFS service",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "webHdfsUrl",
                "type": "STRING"
            }
        },
        {% endif %}

        {% for input_model in input_models %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "The storage type where the input model is stored.",
                "mandatory": true,
                "defaultValue": null,
                "value": "hdfs",
                "key": "dataStorageType-input-model",
                "type": "STRING"
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "description": "HDFS path where to read the model",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "input-model",
                "type": "STRING",
                "inputData": true,
                "outputData": false
            }
        },
        {% endfor %}

        {% for output_model in output_models %}
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "The storage type where the input model is stored.",
                "mandatory": true,
                "defaultValue": null,
                "value": "hdfs",
                "key": "dataStorageType-output-model",
                "type": "STRING"
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "description": "HDFS path where to save the model",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "output-model",
                "type": "STRING",
                "inputData": true,
                "outputData": false
            }
        },
        {% endfor %}
 {
            "it.eng.alidalab.applicationcatalogue.domain.service.ApplicationProperty": {
                "description": "Dataset delimiter",
                "mandatory": false,
                "defaultValue": ",",
                "value": null,
                "key": "delimiter",
                "type": "STRING",
                "inputData": false,
                "outputData": false
            }
        },
        {
             "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                 "description": "Container image pull policy used when pulling images within Kubernetes. Valid values are Always, Never, and IfNotPresent.",
                 "mandatory": true,
                 "defaultValue": "Always",
                 "value": null,
                 "key": "spark.kubernetes.container.image.pullPolicy",
                 "type": "STRING",
                 "externalized": false,
                 "uri": null
             }
         },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Class to use for serializing objects that will be sent over the network or need to be cached in serialized form",
                "mandatory": false,
                "defaultValue": "0",
                "value": null,
                "key": "spark.serializer",
                "type": "INT",
                "minValue": "0",
                "maxValue": "1",
                "measure": "NONE",
                "mappings": [
                    {
                        "intValue": 0,
                        "name": "org.apache.spark.serializer.JavaSerializer"
                    },
                    {
                        "intValue": 1,
                        "name": "org.apache.spark.serializer.KryoSerializer"
                    }
                ],
                "category": true
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "The codec used to compress internal data such as RDD partitions, event log, broadcast variables and shuffle outputs",
                "mandatory": false,
                "defaultValue": "0",
                "value": null,
                "key": "spark.io.compression.codec",
                "type": "INT",
                "minValue": "0",
                "maxValue": "2",
                "measure": "NONE",
                "mappings": [
                    {
                        "intValue": 0,
                        "name": "lz4"
                    },
                    {
                        "intValue": 1,
                        "name": "lzf"
                    },
                    {
                        "intValue": 2,
                        "name": "snappy"
                    }
                ],
                "category": true
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Maximum size of map outputs to fetch simultaneously from each reduce task",
                "mandatory": false,
                "defaultValue": "50331648",
                "value": null,
                "key": "spark.reducer.maxSizeInFlight",
                "type": "INT",
                "minValue": "25165824",
                "maxValue": "100663296",
                "measure": "BYTES",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Default number of partitions in RDDs returned by transformations like join, reduceByKey, and parallelize when not set by user",
                "mandatory": false,
                "defaultValue": "1",
                "value": null,
                "key": "spark.default.parallelism",
                "type": "INT",
                "minValue": "1",
                "maxValue": "16",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Size of each piece of a block for TorrentBroadcastFactory",
                "mandatory": false,
                "defaultValue": "4194304",
                "value": null,
                "key": "spark.broadcast.blockSize",
                "type": "INT",
                "minValue": "2097152",
                "maxValue": "8388608",
                "measure": "BYTES",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Number of cores to allocate for each task",
                "mandatory": false,
                "defaultValue": "1",
                "value": null,
                "key": "spark.task.cpus",
                "type": "INT",
                "minValue": "1",
                "maxValue": "8",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "If set to true, performs speculative execution of tasks",
                "mandatory": false,
                "defaultValue": "0",
                "value": null,
                "key": "spark.speculation",
                "type": "BOOLEAN",
                "minValue": "0",
                "maxValue": "1",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "the spark master",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "spark.master",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "Spark app name",
                "mandatory": true,
                "defaultValue": "{{ name }}",
                "value": null,
                "key": "spark.name",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "Python module",
                "mandatory": true,
                "defaultValue": "main",
                "value": null,
                "key": "pythonModule",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "Number of executors",
                "mandatory": true,
                "defaultValue": 3,
                "value": null,
                "key": "spark.executor.instances",
                "type": "INT",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "Kubernetes secrets used to pull images from private image registries",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "spark.kubernetes.container.image.pullSecrets",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "Container image to use for the Spark application",
                "mandatory": true,
                "defaultValue": {{json.dumps(docker_image.replace("docker://", ""))}},
                "value": null,
                "key": "spark.kubernetes.container.image",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "The namespace that will be used for running the driver and executor pods",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "spark.kubernetes.namespace",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Whether to compress map output files",
                "mandatory": false,
                "defaultValue": "1",
                "value": null,
                "key": "spark.shuffle.compress",
                "type": "BOOLEAN",
                "minValue": "0",
                "maxValue": "1",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.StaticProperty": {
                "description": "The URL for HDFS service",
                "mandatory": true,
                "defaultValue": null,
                "value": null,
                "key": "hdfsUrl",
                "type": "STRING",
                "externalized": false,
                "uri": null
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Whether to compress broadcast variables before sending them",
                "mandatory": false,
                "defaultValue": "1",
                "value": null,
                "key": "spark.broadcast.compress",
                "type": "BOOLEAN",
                "minValue": "0",
                "maxValue": "1",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        },
        {
            "it.eng.alidalab.applicationcatalogue.domain.service.TuningProperty": {
                "description": "Whether to compress data spilled during shuffles",
                "mandatory": false,
                "defaultValue": "1",
                "value": null,
                "key": "spark.shuffle.spill.compress",
                "type": "BOOLEAN",
                "minValue": "0",
                "maxValue": "1",
                "measure": "NONE",
                "mappings": [],
                "category": false
            }
        }
    ],
    "metrics": []
}







class Dataset:
    def __init__(self, name=None, description=None , columns_type=None):
        self.name = name
        self.description = description
        self.columns_type = columns_type
        


def str2bool(v):

    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False



import argparse
from .property import Property
from .type_utils import str2bool
import sys
from .dataset import Dataset
from .model import Model
from .translation_dictionary import translation
import json
import copy
from .generator import generate_meta_model
import re


class ArgumentParser():
    def __init__(self, name, description, mode="batch", area="preparation", framework="default", docker_image=None):
        self.parser = argparse.ArgumentParser()
        self.properties = []
        self.input_datasets = []
        self.output_datasets = []
        self.input_models = []
        self.output_models = []
        
        self.name = name.lower().replace(" ", "-")
        self.description = description
        self.mode = mode
        self.framework = framework
        self.area = area

        if docker_image is not None:
            self.set_docker_image(docker_image)
        else:
            self.docker_image = docker_image


    def add_argument(self, argument, 
                        help="Please fill argument description and upload again your service", 
                        type=str, 
                        default=None,
                        required=False, 
                        action=None,
                        choices=None):


        if action is not None:
            if action == 'store_true':
                type = str2bool
                default = False
            elif action == 'store_false':
                type = str2bool
                default = True
            else:
                print("The action: " + action + " is not currently supported")

        self.properties.append(Property(argument, type=type, help=help, default=default, required=required))
        self.parser.add_argument(argument, help=help, type=type, default=default, required=required, choices=choices)


    def add_input_dataset(self, help=None, columns_type=None):
        self.parser.add_argument("--input_dataset", help=help, required=True, type=str)
        self.input_datasets.append(Dataset(description=help, columns_type=columns_type))

    def add_output_dataset(self, help=None):
        self.parser.add_argument("--output_dataset", help=help, required=True, type=str)
        self.output_datasets.append(Dataset(description=help))

    def add_input_model(self, help=None):
        self.parser.add_argument("--input_model", help=help, required=True, type=str)
        self.input_models.append(Model(description=help))
        self.change_area("analytics")

    def add_output_model(self, help=None):
        self.parser.add_argument("--output_model", help=help, required=True, type=str)
        self.output_models.append(Model(description=help))
        self.change_area("analytics")

    def parse_args(self):
        args, _ = self.parse_known_args()
        return args

    def parse_known_args(self):
        args = sys.argv[1:]

        # Standard python boolean args do not work as in Alida. 
        # Hence, we need to convert boolean args properly.

        # Replace all dashes in the mid of a word with underscores
        args = [replace_dashes(arg) for arg in args]

        alida_compliant_args = []

        # Whenever a --arg_name is found, it is replace with --arg_name=<its-default-value>
        for arg in args:
            # If argument is present in args in the Alida way
            if len(arg.split("="))>1:
                alida_compliant_args.append(arg)
            # If bool is present in the python way put to to alida way
            else:
                default_value = self.get_property_default_value(arg.replace("-", ""))
                if default_value is not None:
                    alida_compliant_args.append(arg + "=" + str(not default_value).lower())

        # If a boolean is not prensent at all in args, add it with its default value
        missing_bool_args = self.find_missing_bool_args(alida_compliant_args)
        for arg in missing_bool_args:
            alida_compliant_args.append(arg + "=" + str(self.get_property_default_value(arg.replace("-", ""))))
        
        return self.parser.parse_known_args(args=alida_compliant_args)


    def get_property_default_value(self, property_name):
        for property in self.properties:
            if property.name == property_name:
                return property.default


    # Find missing boolean arguments
    def find_missing_bool_args(self, args):
        missing_args = []
        for property in self.properties:
            if property.type is str2bool:
                if "--" + property.name + "=false" not in args and "--" + property.name + "=true" not in args and "--" + property.name + "=False" not in args and "--" + property.name + "=True" not in args :
                    missing_args.append("--" + property.name)
        return missing_args


    def generate_json(self, path):
        
        json_data = generate_meta_model(name = self.name, description=self.description, framework=self.framework,
                                        properties = self.properties, 
                                        input_datasets=self.input_datasets,
                                        output_datasets = self.output_datasets,
                                        input_models = self.input_models,
                                        output_models = self.output_models,
                                        docker_image=self.docker_image,
                                        area = self.area,
                                        mode=self.mode
                                        )

        with open(path, "w") as outputFile:
            json.dump(json.loads(clean_json(json_data)), outputFile, indent=4)    #, sort_keys=True)


    def set_docker_image(self, docker_image):
        self.docker_image = "docker://" + docker_image

    # Based on the class extended by the developer to create his service, find out ALIDA service mode
    def set_mode(self, _class):
        self.mode =_class.alida_service_mode
        if self.mode == "source" or self.mode == "sink":
            self.area = "ingestion"

    def change_area(self, new_area):
        self.area = new_area

def clean_json(string):
    string = re.sub(",[ \t\r\n]+}", "}", string)
    string = re.sub(",[ \t\r\n]+\]", "]", string)

    return string

def replace_dashes(string):
    key = string.split("=")[0]
    value = string.replace(key, "")
    return "--" + key.replace("-", "_")[2:] + value
